/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   --------------------------------------------------------------------  *
 *       Handles .BMP bitmap file load/save processing.                    *
 *       Used some .bmp articles from the 'net, describing the format      *
 *                                                                         *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <math.h>
#include "MyGlobal.h"


// here a pointer to the RawImg structures of all loaded .bmp files is stored

typedef struct {
RawImg *a;
char fn[60];
}RawImgFnSv;

// an array of RawImg pointers. The database keeps a pointer to each RawImg created with CreateRaw (used for safe deallocation)
RawImg *PDB[1024];
int PDBCount = 0;

RawImgFnSv BMPSv[16];
int BMPFnSvCnt=0;

void strcpy(char *a, const char *b)
{int i=0;
 do { a[i] = b[i]; } while (b[i++]);
}


void add_pdb(RawImg *a)
{
 PDB[PDBCount++] = a;
}

void del_pdb(RawImg *a)
{int i;
 i = PDBCount - 1;
 while (i>=0 && PDB[i]!=a) i--;
 if (i<0) return;
 while (i<PDBCount - 1)
 	PDB[i] = PDB[i+1];
}

int find_pdb(RawImg *a)
{int i;
 for (i=0;i<PDBCount;i++)
 	if (PDB[i]==a) return 1;
 return 0;
}

/**************************************
 * Internal Proc - Create Raw         *
 *------------------------------------*
 * *a - Pointer to the target RawImg  *
 *  x - desired width                 *
 *  y - desired height                *
 *------------------------------------*
 * On failure, a is NULL and CreateRaw*
 * returns 0.                         *
 **************************************/

int CreateRaw(RawImg *a, int x, int y)
{
 int asize;
 a->x = x;
 a->y = y;
 if (find_pdb(a)) {
 	free(a->data);
	}
 asize = x * y * 4;
 /*if (asize <= 64) */asize += 16;
 a->data = (int*) malloc(asize);
 if (a->data == NULL) {
 	a = NULL;
	del_pdb(a);
	return 0;
 	}
 for (int i=0;i<x*y;i++) a->data[i] = 0;
 return 1;
}


/**********************************
 * Frees the memory allocated for *
 * the data field of the given    *
 * RawImg                         *
 *--------------------------------*
 *  *a - Pointer to a RawImg      *
 *--------------------------------*
 * returns 1 on success, or       *
 * 0 otherwise                    *
 **********************************/
int FreeRaw(RawImg *a)
{
	if (!find_pdb(a)) return 0;
	del_pdb(a);
	free(a->data);
	return 1;
}

/**************************************
 * Proc - Shrinks a to a new RawImg   *
 *        b, with a ratio of 1:2      *
 *------------------------------------*
 *  a - The source RawImg structure   *
 * *b - Pointer to the target         *
 **************************************/
void Shrink(const RawImg a, RawImg *b)
{int i,j,p=0;
 CreateRaw(b, a.x/2, a.y/2);
 for (i=0;i<a.x*a.y/4;i++) b->data[i]=0;
 for (j=0;j<a.y;j++)
 	for (i=0;i<a.x;i++)
		b->data[(i/2)+(j/2)*(a.x/2)] += multiplycolor(a.data[p++], 16384);
 //if (b->x<=4)
 	b->data[b->x*b->y] = b->data[b->x*b->y-1];
/*
	The last "hack" is needed for the SSE version of drawit. If you carefully trace the source, you will see
	that the bilinear filtering version gets 2 pixels with one `movq', which is not identical to the
	P5 code, which gets pixels one by one and mods the index of the pixel in the texture by the texture size (so
	no texel outside of the texture is taken). In the SSE version the second pixel is not checked for being
	outside the texture (but it can be just the pixel behind the last in the texture memory), so we copy the
	last pixel to the position after it. This is not correct, but does the job.
*/
}

/**************************************
 * Function - Loads a .bmp file into  *
 *            a RawImg                *
 * Supports 8-,24- and 32-bit images  *
 *------------------------------------*
 * fn - The path to the file.         *
 * *a - The target RawImg structure   *
 *------------------------------------*
 * On failure, LoadBmp returns 0.     *
 **************************************/
int LoadBmp(const char *fn, RawImg *a)
{FILE *f;
 int i, j, p, k;
 BMHEADER hd;
 BMiHEADER hi;
 int pal[256];
 int toread = 0;
 unsigned char *xx;
 int rowsz;
 unsigned short sign;

 if ((f = fopen(fn, "rb"))==NULL) {
 	printf("LoadBmp:Can't open file: `%s'\n", fn);
 	return 0;
	}
 if (!fread(&sign, 2, 1, f)) return 0;
 if (sign!=BM_MAGIC) { printf("LoadBmp:`%s' is not a BMP file.\n", fn); return 0; }
 if (!fread(&hd, sizeof(hd), 1, f)) return 0;
 if (!fread(&hi, sizeof(hi), 1, f)) return 0;

 /* header correctness checks */
 if (!(hi.bitsperpixel==8 || hi.bitsperpixel==24 || hi.bitsperpixel==32))
 	{printf("LoadBmp:Cannot handle file format at %d bpp.\n",hi.bitsperpixel); return 0 ; }
 if (hi.channels!=1) {
 	printf("LoadBmp: cannot load multichannel .bmp!\n");
	return 0;
 	}
 /* ****** header is OK *******/

 // if image is 8 bits per pixel or less (indexed mode), read some pallete data
 if (hi.bitsperpixel <= 8) {
 	toread = (1<<hi.bitsperpixel);
 	if (hi.colors) toread=hi.colors;
 	if (!fread(pal, toread, 4, f)) return 0;
	}
 toread = hd.bfImgOffset - (54 + toread*4);
 if (toread)
 // skip the rest of the header
 	for (i=0;i<toread;i++)
		fgetc(f);
 k = hi.bitsperpixel / 8;
 rowsz = hi.x * k;
 if (rowsz%4)
 	rowsz = (rowsz / 4 + 1) * 4; // round the row size to the next exact multiple of 4
 xx = (unsigned char *) malloc(rowsz);
 p = hi.x * hi.y;
 if (!CreateRaw(a, hi.x, hi.y)) {
 	printf("LoadBmp: cannot allocate memory for bitmap! Check file integrity!\n");
	fflush(stdout);
	free(xx);
	fclose(f);
	return 0;
 	}
#ifdef DEBUG
 printf("LoadBmp: Loading \"%s\":%dx%dx%dbit BMP...", fn, hi.x, hi.y, hi.bitsperpixel); fflush(stdout);
#endif
 for (j=hi.y-1;j>=0;j--) {// bitmaps are saved in inverted y
	if (!fread(xx, 1, rowsz, f)) {
		printf("LoadBmp: short read while opening `%s', file is probably incomplete!\n", fn);
		free(xx);
		fclose(f);
		return 0;
		}
	p -= hi.x;
	for (i=0;i<hi.x;i++){ // actually read the pixels
		if (hi.bitsperpixel>8)
			a->data[p++] = xx[i*k] + ((int) xx[i*k+1])*256 + ((int) xx[i*k+2])*65536;
			else
			a->data[p++] = pal[xx[i*k]];
		}
	p -= hi.x;
 	}
 fclose(f);
 free(xx);
#ifdef DEBUG
 printf("OK\n");
 fflush(stdout);
#endif
 BMPSv[BMPFnSvCnt].a = a;
 strcpy(BMPSv[BMPFnSvCnt++].fn, fn);
 return 1;
}

/**************************************
 * Function - Writes a .bmp file from *
 *            a RawImg                *
 *------------------------------------*
 * fn - The path to the file.         *
 * *a - The source RawImg structure   *
 *------------------------------------*
 * On failure, SaveBmp returns 0.     *
 **************************************/
int SaveBmp(const RawImg a, const char *fn)
{FILE *f;
 int i,j, rowsz;
 unsigned short msh=BM_MAGIC;
 BMHEADER hd;
 BMiHEADER hi;
 char xx[RES_MAXX*3];


 if ((f=fopen(fn, "wb"))==NULL) {
	printf("SaveBmp:can't open `%s' for writing, check permissions!\n", fn);
	return 0;
	}
 // fill in the header:
 rowsz = a.x * 3;
 if (rowsz % 4)
 	rowsz += 4 - (rowsz % 4); // each row in of the image should be filled with zeroes to the next multiple-of-four boundary
 hd.fs = rowsz*a.y + 54; //std image size :)
 hd.lzero = 0;
 hd.bfImgOffset = 54;
 hi.ihdrsize = 40;
 hi.x = a.x; hi.y = a.y;
 hi.channels = 1;
 hi.bitsperpixel = 24; //RGB format
 // set the rest of the header to default values:
 hi.compression = hi.biSizeImage = 0;
 hi.pixPerMeterX = hi.pixPerMeterY = 0;
 hi.colors = hi.colorsImportant = 0;

 fwrite(&msh, 2, 1, f); // write 'BM'
 fwrite(&hd, sizeof(hd), 1, f); // write file header
 fwrite(&hi, sizeof(hi), 1, f); // write image header
 for (j=a.y-1;j>=0;j--) {
 	for (i=0;i<a.x;i++) {
		xx[i*3  ] = (0xff     & a.data[j*a.x+i]);
		xx[i*3+1] = (0xff00   & a.data[j*a.x+i]) >> 8;
		xx[i*3+2] = (0xff0000 & a.data[j*a.x+i]) >> 16;
		}
        fwrite(xx, rowsz, 1, f);
 	}
 fclose(f);
 return 1;
}

void Take_Snapshot(const RawImg a)
{int i=1;
 char fn[30];
 FILE *f;

 while (42) {
	sprintf(fn, "fract_snapshot%02d.bmp", i);
	f = fopen(fn, "rb");
	if (f==NULL) break;
	   else fclose(f);
	i++;
 	}
 SaveBmp(a, fn);
}

static char outbuff[60];

char *getTexFn(RawImg *x)
{int i, found;
 for (found=0,i=0;i<BMPFnSvCnt;i++)
 	if (BMPSv[i].a == x) {
		strcpy(outbuff, BMPSv[i].fn);
		found = 1;
		break;
		}
 if (!found) sprintf(outbuff, "error");
 return outbuff;
}

void bitmap_close(void)
{int i;
 for (i=0;i<PDBCount;i++) {
 	printf("bitmap_close: freeing a %dx%d bitmap...\n", PDB[i]->x, PDB[i]->y);
  	free(PDB[i]->data);
	}
}
