/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *  render.cpp - all rendering routines, seperated for non-SSE and SSE-    *
 *               enabled CPU's                                             *
 *                                                                         *
 ***************************************************************************/

#include <SDL/SDL_thread.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "MyGlobal.h"
#include "cross_vars.h"


#include "fractasm.h"

Uint32 floorbuffer[RES_MAXX*RES_MAXY];
unsigned short fbuffer[RES_MAXX*RES_MAXY];
int RowInfo[RES_MAXY];
double Area_const = 1.95;
int UseThreads = -1;

/* integrated lighting and Multiplication routine.
 * this routine replaces the following code:
 *
 *	screencolor = multiplycolor(texturecolor, lform(Lx, Ly, lx, ly))
 * being used somewhere in DrawIt.
 *
 * uses an integrated assembly code in fractasm.h (using SSE) if the required instruction set is available
 */
Uint32 integrated( Uint32 texturecolor, int x, int y, int lx, int ly)
{
 if ( sse_enabled )  igrated(texturecolor, x, y, lx, ly);
 		else multiplycolor(texturecolor, lform(x, y, lx, ly));
}

/* This is called by the raytracer - gets a texture color at coords (x, 0, y) */
Uint32 texture_handle(int x, int y)
{
 integrated(T[TEX_S].data[((x & 255)>>TEX_S) + /*y*/(((y & 255)>>TEX_S)<<(8-TEX_S))],x, y, lx, lz);
}

/*
 *	This handles camera movement using current direction vector of the camera and the supplied
 *	`angle' param (angle = 0 means "move straight ahead"; angle = -M_PI means 'move backwards' and so on.
 */
void move_camera(double angle)
{double a, b, x[3];
 int i;
 a = alpha + angle;
 b = beta;
 if (angle == M_PI/2 || angle == -M_PI/2) b = 0;
 if (angle == M_PI)	b = -b;
 x[0] = sin(a)*cos(b);
 x[1] = sin(b);
 x[2] = cos(a)*cos(b);
 for (i=0;i<3;i++)
 	cur[i]+=(x[i]*delta*speed);
 camera_moved = 1;
}

/*
 *	Rotates the camera vector, depending on last frame generation time and mouse sensitivity
 */
void rotate_camera(double alphai, double betai)
{
 alpha += (alphai*delta*anglespeed);
 beta +=  (betai *delta*anglespeed);
}

void check_params(void)
{
 if (cur[1]<daFloor+1) cur[1] = daFloor+1;
 if (cur[1]>daCeiling -1) cur[1] = daCeiling -1;
 if (beta<-M_PI/2) beta = -M_PI/2;
 if (beta> M_PI/2) beta =  M_PI/2;
}

double czero(double x)
{
 if (x<0.0001) x = 9999;
 return x;
}

/*
 *	This is the function to cope with the blur effect (if it is enabled in MyGlobal.h) - the idea used here is
 *	to generate 8 consecutive frames and merge them in a single frame which is copied to the real framebuffer and
 *	displayed as usual. We are using src for temporary (hidden) framebuffer, `bb' to accumulate result from the
 *	eight frames. `fb' is the `real' framebuffer. Note that pixels are right shifted by 3 before added to the
 *	accumulation buffer (this does serious image degradation, since output image is 15 bit).
 *
 *	A MMX version of this routine exists in fractasm.h
 */

void blur_do(Uint32 *bb, Uint32 *src, Uint32 *fb, int count, int shouldcopy)
{
	int i;
	if (mmx_enabled)
		blur_do_MMX(bb, src, fb, count, shouldcopy);
		else {
			for(i=0;i<xres()*yres();i++)
				bb[i]+= ((src[i] >> 3) & 0x1f1f1f1f);
			if (shouldcopy) {
				memcpy(fb, bb, count*4);
				for (i=0;i<count;i++)
					bb[i]=0;
 			}

		}
}

/*
 *	This code deals with post-frame generation stuff, which is common for the P5 and SSE version of DrawIt
 */

void postframe_do(SDL_Surface *p, SDL_Overlay *ov, Uint32 *fb)
{
 static char buff[1024];
 SDL_Rect ddestrect;

 ddestrect.x = 0;
 ddestrect.y = 0;
 ddestrect.w = xres();
 ddestrect.h = yres();
	prof_enter(PROF_ANTIALIAS);
#ifdef BLUR
 antialias(blurbuffer);
#else
 antialias(fb);
#endif
	prof_leave(PROF_ANTIALIAS);
	prof_enter(PROF_SHADER);
#ifdef BLUR
 //	shader_FFT_Filter(blurbuffer, blurbuffer, xres(), yres());
    if (use_shader)
 	//shader_edge_glow(blurbuffer, blurbuffer, xres(), yres());
#else
// 	shader_FFT_Filter(fb, fb, xres(), yres());
    if (use_shader)
 //	shader_edge_glow(fb, fb, xres(), yres());
	//shader_blur(fb, fb, xres(), yres(), 7);
#endif
 	prof_leave(PROF_SHADER);
#ifdef MAKE_CHECKSUM
	prof_enter(PROF_CHECKSUM);
 for (i=0;i<xres()*yres();i++) {
 	cksum += (fb[i]&0xffff)^(fb[i]>>16);
	}
	prof_leave(PROF_CHECKSUM);
#endif
#ifdef BLUR
	prof_enter(PROF_BLUR_DO);
 blur_do(bb1, blurbuffer, fb, xres()*yres(), (vframe%8==0));
 	prof_leave(PROF_BLUR_DO);
#endif
#ifdef SHOWFPS
	prof_enter(PROF_SHOWFPS);
 if (developer) {
 	fpsfrm++;
 	if (SDL_GetTicks()-fpsclk>FPS_UPDATE_INTERVAL || not_fps_written_yet) {
 		not_fps_written_yet = 0;
 		sprintf(buff, "fps: %.1lf", (double) fpsfrm / ((SDL_GetTicks()-fpsclk)/1000.0));
		printxy(p, fb, font0, 0, 0, FPS_COLOR, 0.8, buff);
 		fpsfrm = 0;
 		fpsclk = SDL_GetTicks();
 		} else printxy(p, fb, font0, 0, 0, FPS_COLOR, 0.8, buff);
 }
 	prof_leave(PROF_SHOWFPS);
#endif
 p = get_screen();
 if (ov==NULL) { // use the surface directly
 	Slock(p);
		prof_enter(PROF_MEMCPY);
 	memcpy(p->pixels, fb, p->h*p->w*4);
		prof_leave(PROF_MEMCPY);
 	Sulock(p);
#ifdef ACTUALLYDISPLAY
		prof_enter(PROF_SDL_FLIP);
 	SDL_Flip(p);
		prof_leave(PROF_SDL_FLIP);
#endif
 	}
	else { // use the overlay
	SDL_LockYUVOverlay(ov);
		prof_enter(PROF_CONVERTRGB2YUV);
	ConvertRGB2YUV((Uint32 *) ov->pixels[0], fb, p->h*p->w);
		prof_leave(PROF_CONVERTRGB2YUV);
	SDL_UnlockYUVOverlay(ov);
#ifdef ACTUALLYDISPLAY
		prof_enter(PROF_DISPLAYOVERLAY);
	SDL_DisplayYUVOverlay(ov, &ddestrect);
		prof_leave(PROF_DISPLAYOVERLAY);
#endif
	}
 vframe++;
#ifdef FSTAT
 printf("frame %6d: good/bad/realbad = %7.3lf%%/%7.3lf%%/%7.3lf%%\n", vframe,
 (double) 100.0*good    / ((double) xres()*yres()),
 (double) 100.0*worse   / ((double) xres()*yres()),
 (double) 100.0*realbad / ((double) xres()*yres())); fflush(stdout);
#endif
}

/*
 *	This routine does the common stuff, identical for both the P5 and SSE version of Drawit
 */

void preframe_do(SDL_Surface *p, Uint32 *ptr, double *lw)
{
 int i, j;
 double btm, ttm;
 double obex[3];
 	prof_enter(PROF_PREFRAME_MEMSET);
 memset(ptr, 0, xsize_render(p->w)*ysize_render(p->h)*4);
 	prof_leave(PROF_PREFRAME_MEMSET);
 // preframe things -----------------------------------
 	prof_enter(PROF_SPHERE_THINGS);
 btm = bTime();
 if (!(vframe&cd_frames)) { // first frame
 	for (i=0;i<spherecount;i++)
		if (sp[i].flags & ANIMATED) {
			calculate_XYZ(sp+i, btm, obex);
			sp[i].x = obex[0];
			sp[i].y = obex[1];
			sp[i].z = obex[2];
			sp[i].xi = 0;
			sp[i].yi = 0;
			sp[i].zi = 0;
			sp[i].time = btm;

			}
    camera_moved = 1;
 	}
 if (vframe%cd_frames) // every frame but the first
 for (i=0;i<spherecount;i++) {
 	if (sp[i].flags & ANIMATED) {
		calculate_XYZ(sp+i, btm, obex);
		sp[i].xi = (obex[0] - sp[i].x) / (btm - sp[i].time);
		sp[i].yi = (obex[1] - sp[i].y) / (btm - sp[i].time);
		sp[i].zi = (obex[2] - sp[i].z) / (btm - sp[i].time);
		camera_moved = 1;
		}
	}
	prof_leave(PROF_SPHERE_THINGS);
	prof_enter(PROF_PHYSICS);
 if (Physics) { // should we process physics?
 	camera_moved = 1;
 	for (i=0;i<spherecount;i++)
		apply_gravity(sp+i, btm); // first of all, Gravity
	for (i=0;i<spherecount;i++)
		apply_air(sp+i, btm, APPLY_ALL, 1);    // air resistance
	if (CollDetect)			 // do we process collisions?
	/* The algorythm is simple:
		We iterate thru all collideable spheres and check for collision.
		If a collision occurs, we update vectors (and other data), and set the collided balls'
		time to the moment of the collision. The balls are ready to collide again within the same
		time segment of the frame, but now they differ from the others, being advanced a little
		in the future (this is very important). After everything is completed, all balls' times
		are synced to the new time
	*/
		for (j=0;j<spherecount;j++) if (sp[j].flags & COLLIDEABLE)
			for (i=j+1;i<spherecount;i++) if (sp[i].flags & COLLIDEABLE)
				if ((ttm=collide(sp+i, sp+j, btm))>0) // collision detected
					ProcessIncident(sp+i, sp+j, ttm);
	/*
		Here is where advance takes place. All uncollided (and some collided, too)
		spheres are moved with respect to the remaining time to the sync time and
		their movement vectors, too.
	*/
	for (i=0;i<spherecount;i++)
		advance(sp+i, btm);
 	}
	prof_leave(PROF_PHYSICS);
	prof_enter(PROF_SORT);
 sort_fract_array(sp, spherecount);
 	prof_leave(PROF_SORT);

	prof_enter(PROF_PROJECT);
for (i=0;i<spherecount;i++) {
 	projectIt(sp+i, pp[i], num_sides+i, ptr, p, cur, w, xsize_render(xres()), ysize_render(yres()), i+1);
	pdlo[i] = pith0(sp[i].x, sp[i].y, sp[i].z, lx, ly, lz);
 	sp[i].dist = pith0(sp[i].x, sp[i].y, sp[i].z, cur[0], cur[1], cur[2]);
	}
	prof_leave(PROF_PROJECT);
	prof_enter(PROF_PASS);
 PassPre(pdlo, pith(lw, cur), lw);
 	prof_leave(PROF_PASS);
}

/*
 *	Parameter initialisation routine - initialises the vector grid basics and interpolation increasers
 */
void bash_preframe(double *lw, double *tt, double *ti, double *tti)
{
 int i;
 lw[0] = lx; lw[1] = ly; lw[2] = lz;
 calc_grid_basics(cur, alpha, beta, w);
 for (i=0;i<3;i++) {
 	tt[i]=w[0][i];
	ti[i]=(w[1][i]-w[0][i])/((double) xsize_render(xres()));
	tti[i]=(w[2][i]-w[0][i])/((double) ysize_render(yres()));
 	}
}

/*
 *	Chooses which texture instance to use based on the distance of the row from the camera and the angle
 *	the ray from the camera encloses with that row.
 */
void choose_texture(double *cur, int xr, int yr, double cx, double cxi, double cy, double cyi, double hx, double hy,
			double dp, int *xtex, int *xandmask, int *xshl, int *tam, int *notmask
			)
{
	double parea; // parea is the area (in square texels) which single ray covers
	parea = sqrt(cxi * cxi + cyi * cyi) * sqrt(sqr(cx - hx) + sqr(cy - hy));
	*xtex = 0; *xandmask = MAX_TEXTURE_SIZE; *xshl = (int) log2(*xandmask);
	/*
		The idea is simple. Since we want to bilinear filter (or do nearest neighbour) all the time
		we do want a single ray to cover an area of approx 1 sq texel, we get the area a ray will have
		on the 256x256 texture. If the area is above `Area_const' (an meaningful value [1..4]) we want
		to get a smaller texture - a 128x128 version mipmap will have 4 times less texels, resulting in
		four times less area. So we get lower texture sizes and divide the area by 4 each time, until
		the area is acceptable or we reached using 1x1 texture.
	*/
	while (parea>Area_const && *xtex<end_tex) {
		(*xtex)++; (*xandmask)/=2; (*xshl)--;
		parea /= 4.0;
		}
	(*xandmask)--;
	*notmask = (0xffffffff >> (LOG2_TEXTURE_AREA+*xtex)) << (LOG2_TEXTURE_AREA+*xtex);
	*tam = (1 << (2*(*xshl))) - 1;
}

/*************************************************
 * This is the non-SSE version of the renderer,  *
 * which is single-threaded and does all-in-one: *
 * pre-filling, sphere and floor rendering. Uses *
 * function calls, FPU and integer arithmetic    *
 *************************************************/

void DrawIt_P5(SDL_Surface *p, SDL_Overlay *ov)
{int ii,i,j,k;
#ifdef FSTAT
 int good=0, worse=0, realbad=0;
#endif
 int Lx=99999, Ly=99999, Lr=0, SavergX, SavergY;
 int Bx, By, notmask, Tc, boffset, texandmask;
 int Fx, Fy;
 Uint32 *ptr;
 double t[3], tt[3], ti[3], tti[3], v[3], lw[3], ydist, scalefactor, dp, ut[3];

 int xtex, xandmask, xshl;

 float cx, cy, cxi, cyi, hx, hy;

 int gX, gY, gxi, gyi, xr, yr;

 Uint32 fcol=0, bfcol;
 double opacity,bopacity;
 int dropped, backd, foundsphere, addstuff;

 xr = xsize_render(xres());
 yr = ysize_render(yres());

 bash_preframe(lw, tt, ti, tti);

#ifdef BLUR
 ptr = (Uint32 *)blurbuffer;
#else
 ptr = (Uint32 *) framebuffer;
#endif
 preframe_do(p, ptr, lw);

 for (j=0;j<yr;j++) {
 	for (k=0;k<3;k++) {
		t[k]=tt[k];
		ut[k] = tt[k] - tti[k];
	}
	// check for horizon
	if ((ydist=vfabs(cur[1]-t[1]))>DST_THRESHOLD) {
		dp = (cur[1]>t[1]?(cur[1]-daFloor):(daCeiling-cur[1]));
		scalefactor = dp/ydist;
		cx = cur[0] + (t[0]-cur[0])*scalefactor;
		cy = cur[2] + (t[2]-cur[2])*scalefactor;
		cxi = (cur[0]+(t[0]+ti[0]*xr-cur[0])*scalefactor - cx)/((double) xr);
		cyi = (cur[2]+(t[2]+ti[2]*yr-cur[2])*scalefactor - cy)/((double) yr);
		// calculate second "helping" point from the previos image row
		scalefactor = (cur[1]>ut[1]?(cur[1]-daFloor):(daCeiling-cur[1])) / vfabs(cur[1]-ut[1]);
		hx = cur[0] + (ut[0]-cur[0])*scalefactor;
		hy = cur[2] + (ut[2]-cur[2])*scalefactor;

		gX = (int) (cx*65536.0);
		gY = (int) (cy*65536.0);
		gxi =(int) (cxi*65536.0);
		gyi =(int) (cyi*65536.0);
		// choose apropriate texture from the T array:
		choose_texture(cur, xr, yr, cx, cxi, cy, cyi, hx, hy, dp,
			&xtex, &xandmask, &xshl, &texandmask, &notmask);
		for (i=0;i<xr;i++)
			{
			addstuff = 0;
			dropped = *ptr;
			if (dropped) {
				backd = dropped>>16;
				dropped &= 0xffff;
				v[0] = t[0]-cur[0]; v[1] = t[1] - cur[1]; v[2] = t[2] - cur[2];
				norm(v);
				if (testsphere(sp+dropped-1, v, cur)) {
					fcol = Solve3D(sp + dropped-1 , v, cur, lw, &opacity, 0);
					if (opacity > 0.99)  *ptr=fcol;
					               else {
						       	if (backd) {
						       		if (testsphere(sp+backd-1, v, cur)) {
									bfcol = Solve3D(sp + backd - 1, v, cur, lw, &bopacity, 0);
									if (bopacity>0.99) *ptr = blend(fcol, bfcol, opacity);
										else	 {
										fcol = blend_p5(fcol, bfcol, opacity);
										opacity = opacity + (bopacity*(1-opacity));
										addstuff=1;
										dropped=0;
										}
						       			}
									else {*ptr=fcol; addstuff=1; dropped = 0;}
								}
							else{addstuff = 1; dropped = 0;}
						       }
#ifdef FSTAT
					good++;
#endif
					}
					else {// too bad - we'll have to iterate all the spheres to find the right one...
					foundsphere = 0;
#ifdef FSTAT
					worse++;
#endif
					for (ii=spherecount-1;ii>=0;ii--) if (testsphere(sp + ii, v, cur)) { // sphere found...
						fcol = Solve3D(sp + ii, v, cur, lw, &opacity, 0);
						if (opacity > 0.99) *ptr=fcol;
						               else {addstuff = 1; dropped = 0; }
						foundsphere = 1;
						break;
						}
					if (!foundsphere)
						{
						/******************************************************
						 * If we reached here, this means that the pre-filler *
						 * misled us there is a sphere... we've done alot a   *
						 * work to make it here, so the only thing that keeps *
						 * our engine from renderering *REALLY* slow is that  *
						 * such BAD pixels are very rare (about 1-2 %)        *
						 ******************************************************/
#ifdef FSTAT
						 realbad++;
#endif
						dropped = 0; addstuff=0;
						}
					}
				}
			if (!dropped) {
				if (!bilfilter) {
					SavergX = r(gX); SavergY = r(gY);
					if (SavergX==Lx && SavergY==Ly)
						*(ptr) = Lr;
						else{
						Lx=SavergX; Ly=SavergY;
						*(ptr) = Lr = multiplycolor(T[xtex].data[(((unsigned) (Lx)>>xtex) & xandmask) + (((((unsigned)(Ly)>>xtex) & xandmask))<<xshl)],lform(Lx, Ly, lx, lz));
						}
					}
					else {
					Bx = (gX<0?(notmask|(gX>>(16+xtex))):(gX>>(16+xtex)));
					By = (gY<0?(notmask|(gY>>(16+xtex))):(gY>>(16+xtex)));
					Fx = ((gX>>xtex)&0xffff);
					Fy = ((gY>>xtex)&0xffff);
					boffset = ((Bx & xandmask)) + (((By & xandmask))<<xshl);
					Tc = multiplycolor(T[xtex].data[boffset				], i2im(65536-Fx,65536-Fy))+
					multiplycolor(T[xtex].data[(boffset+1) &texandmask		], i2im(Fx  ,65536-Fy))+
					multiplycolor(T[xtex].data[(boffset+xandmask+1)&texandmask	], i2im(65536-Fx,  Fy))+
					multiplycolor(T[xtex].data[(boffset+xandmask+2)&texandmask	], i2im(Fx  ,  Fy));
					*(ptr) = multiplycolor(Tc, lform(r(gX), r(gY), lx, lz));
					}
				}
			if (addstuff) *ptr = blend_p5(fcol, *ptr, opacity);
			// continue interpolation
			ptr++;
			gX += gxi; gY+=gyi;
			t[0]+=ti[0]; t[2]+=ti[2];
			}
		}
		else {
 		for (i=0;i<xr;i++)
			{
			addstuff = 0;
			dropped = (*ptr);
			if (dropped) {
				backd = dropped>>16;
				dropped &= 0xffff;
				v[0] = t[0]-cur[0]; v[1] = t[1] - cur[1]; v[2] = t[2] - cur[2];
				norm(v);
				if (testsphere(sp+dropped-1, v, cur)) {
					fcol = Solve3D(sp + dropped-1 , v, cur, lw, &opacity, 0);
					if (opacity > 0.99)  *ptr=fcol;
						else{
						       	if (backd) {
						       		if (testsphere(sp+backd-1, v, cur)) {
									bfcol = Solve3D(sp + backd - 1, v, cur, lw, &bopacity, 0);
									if (bopacity>0.99) *ptr = blend_p5(fcol, bfcol, opacity);
										else {
										fcol = blend_p5(fcol, bfcol, opacity);
										opacity = opacity + (bopacity*(1-opacity));
										addstuff=1;
										dropped=0;
										}
						       			}
									else {*ptr=fcol; addstuff=1; dropped = 0;}
								}
							else{addstuff = 1; dropped = 0;}
						       }
#ifdef FSTAT
					good++;
#endif
					}
					else {// too bad - we'll have to iterate all the spheres to find the right one...
#ifdef FSTAT
					worse++;
#endif
					foundsphere = 0;
					for (ii=spherecount-1;ii>=0;ii--) if (testsphere(sp + ii, v, cur)) { // sphere found...
						fcol = Solve3D(sp + ii, v, cur, lw, &opacity, 0);
						if (opacity > 0.99) *ptr=fcol;
						               else {addstuff = 1; dropped = 0; }
						foundsphere = 1;
						break;
						}
					if (!foundsphere)
						{
#ifdef FSTAT
						 realbad++;
#endif
						 dropped = 0; addstuff=0;
						}
					}
				}
			if (addstuff) *ptr = multiplycolorf(fcol, opacity);
			ptr++;
			gX += gxi; gY+=gyi;
			t[0]+=ti[0]; t[2]+=ti[2];
			}
		}
	for (k=0;k<3;k++) tt[k]+=tti[k];
 	}
// frame generation finished:
 postframe_do(p, ov, framebuffer);
}

/*
	render_floor:

	This is a part of the SSE version of the renderer. This creates an image of the floor (and the ceiling)
	only. Shadow-casting is not performed here. Uses assembly-optimized code, with loop unrolling and not
	a single IF in the whole routine :)))

 */

void render_floor(SDL_Surface *p, SDL_Overlay *ov, Uint32 *ptr, int xr, int yr, double *tt, double *ti, double *tti)
{
	int i, j, k;
	float fx[4], fy[4];
	float fxi[4];
	float fyi[4];
	float fxyi[8] __attribute__ ((aligned (16)));
	float lightxy1[12] __attribute__ ((aligned (16)));
	// these alignments are needed to use movaps intead of movups.
	float c64k[4] __attribute__ ((aligned (16))) =
		{BRIGHTNESS_C, BRIGHTNESS_C, BRIGHTNESS_C, BRIGHTNESS_C};
	int gx[4]  __attribute__ ((aligned (8)));
	int gy[4]  __attribute__ ((aligned (8)));
	int gxx[8] __attribute__ ((aligned (8)));
	int save[4] __attribute__ ((aligned (8)));
//	int ttexandmask[2] __attribute__ ((aligned (8)));
	unsigned short ffs[8] __attribute__ ((aligned (8))) =
		{0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x0, 0xffff, 0x0};
	double ydist, dp, scalefactor, cx, cy, cxi, cyi, ut[3], hx, hy;
	int xtex, xandmask, xshl, texandmask, notmask;
	int xxandmask[2];
	Uint32 *ptex;

	for (j=0;j<yr;j++) {
		for (k=0;k<3;k++) ut[k] = tt[k] - tti[k];
		if ((ydist=vfabs(cur[1]-tt[1]))>DST_THRESHOLD) {
				prof_enter(PROF_INIT_PER_ROW);
			dp = (cur[1]>tt[1]?(cur[1]-daFloor):(daCeiling-cur[1]));
			scalefactor = dp/ydist;
			cx = cur[0] + (tt[0]-cur[0])*scalefactor;
			cy = cur[2] + (tt[2]-cur[2])*scalefactor;
			cxi = (cur[0]+(tt[0]+ti[0]*xr-cur[0])*scalefactor - cx)/((double) xr);
			cyi = (cur[2]+(tt[2]+ti[2]*yr-cur[2])*scalefactor - cy)/((double) yr);
			// calculate second "helping" point from the previos image row
			scalefactor = (cur[1]>ut[1]?(cur[1]-daFloor):(daCeiling-cur[1])) / vfabs(cur[1]-ut[1]);
			hx = cur[0] + (ut[0]-cur[0])*scalefactor;
			hy = cur[2] + (ut[2]-cur[2])*scalefactor;

			for (i=0;i<4;i++) {
				gxx[i] = (int) (cxi*262144.0);
				gxx[i+4] = (int) (cyi*262144.0);
				gx[i] = (int) (cx*65536.0) + (int) (i*cxi*65536.0);
				gy[i] = (int) (cy*65536.0) + (int) (i*cyi*65536.0);
				fx[i] = cx + i*cxi;
				fy[i] = cy + i*cyi;
				fxi[i] = cxi*4;
				fyi[i] = cyi*4;
				lightxy1[i] = lx;
				lightxy1[i+4] = lz;
				lightxy1[i+8] = 1.0;
				fxyi[  i] = fxi[i];
				fxyi[4+i] = fyi[i];
			}
			// choose apropriate texture from the T array:
			choose_texture(cur, xr, yr, cx, cxi, cy, cyi, hx, hy, dp,
				&xtex, &xandmask, &xshl, &texandmask, &notmask);
			xxandmask[0] = xxandmask[1] = xandmask;
			// actual rendering follows:
			ptex = (Uint32 *) T[xtex].data;
			i = xr/4;
				prof_leave(PROF_INIT_PER_ROW);
				prof_enter(PROF_WORK_PER_ROW);
			// load some useful values:
			__asm __volatile (
			"	movups	%0,	%%xmm0\n"
			"	movups	%1,	%%xmm1\n"
			"	movups	%2,	%%xmm2\n"
			"	movups	%3,	%%xmm3\n"
			"	movups	%6,	%%xmm7\n"

			"	movq	%4,	%%mm0\n"
			"	movq	8%4,	%%mm1\n"
			"	movq	%5,	%%mm2\n"
			"	movq	8%5,	%%mm3\n"
			"	pxor	%%mm6,	%%mm6\n"
			"	movd	%7,	%%mm6\n"
			:
			:"m"(*fx),"m"(*fy),"m"(*fxi),"m"(*fyi),"m"(*gx),"m"(*gy),"m"(*c64k), "m"(xtex)
			:"memory"
			);
			if (bilfilter) {
				while (i--) {
					__asm __volatile (
				// calculate light degradation:
					"	movaps	%5,	%%xmm4\n" // load light x in xmm4
					"	movaps	0x10%5,	%%xmm5\n" // load light y in xmm5
					"	movaps	%%xmm7,	%%xmm6\n"
					"	subps	%%xmm0,	%%xmm4\n" // xmm4 = [ -current x + lightx ]   x 4
					"	subps	%%xmm1,	%%xmm5\n" // xmm5 = [ -current y + lighty ]   x 4
					"	mulps	%%xmm4,	%%xmm4\n" // xmm4 = [(-current x + lightx)^2] x 4
					"	mulps	%%xmm5,	%%xmm5\n" // xmm5 = [(-current y + lighty)^2] x 4
					"	addps	0x20%5,	%%xmm4\n" // xmm4 = xmm4 + (1, 1, 1, 1)
					"	addps	%%xmm5,	%%xmm4\n" // xmm4 = xmm4 + xmm5 = dist^2

					"	rsqrtps	%%xmm4,	%%xmm4\n" // xmm4 = 1 / dist
					"	mulps	%%xmm4,	%%xmm6\n" // xmm6 = const / dist
					"	rsqrtps %%xmm4, %%xmm4\n" // xmm4 = sqrt(dist)
					"	mulps	%%xmm4,	%%xmm6\n" // xmm6 = const / sqrt(dist)

				// save coordinates from mmx registers to memory:
					"	movq	%%mm0,	%1\n"
					"	movq	%%mm1,	8%1\n"
					"	movq	%%mm2,	%2\n"
					"	movq	%%mm3,	8%2\n"

				// shift all coords xshl positions, arithmetical
					"	psrad	%%mm6,	%%mm0\n"
					"	psrad	%%mm6,	%%mm1\n"
					"	psrad	%%mm6,	%%mm2\n"
					"	psrad	%%mm6,	%%mm3\n"
					"	movq	%%mm0,	%%mm4\n"
					"	movq	%%mm2,	%%mm5\n"
				// save 3rd and 4th x and y coordinate to work on later...
					"	movq	%%mm1,	%3\n"
					"	movq	%%mm3,	8%3\n"
				// shift the first copy additional 16 bits, arithmetical
					"	psrad	$16,	%%mm0\n"
					"	psrad	$16,	%%mm2\n"

	/* situation here:                                          ************
		mm0 = gX1&gX2 >> (16 + xshl) = Bx                 **    ****    **
		mm1 = unused & saved                             *     *  **      *
		mm2 = gY1&gY2 >> (16 + xshl) = By               *     *   **       *
		mm3 = unused & saved                            *         **       *
		mm4 = gX1&gX2 >> (xshl)                         *         **       *
		mm5 = gY1&gY2 >> (xshl)                         *         **       *
		mm6 = xshl                                       *        **      *
		mm7 = xxandmask                                   **   *******  **
                                                                    ************
	*/

					"	pslld	$16,	%%mm4\n"
					"	pslld	$16,	%%mm5\n"
					"	psrld	$17,	%%mm4\n" //mm4 &= 0xffff = FX
					"	psrld	$17,	%%mm5\n" //mm5 &= 0xffff = FY
					"	packssdw %%mm5,	%%mm4\n" //mm4 = (lo to hi) -> [ fx0 fx1 fy0 fy1 ]
					"	psllw	$1,	%%mm4\n"

					"	pand	%8,	%%mm2\n"
					"	pand	%8,	%%mm0\n" // mm0 &= xandmask
					SHIFT_LEFT_LOG2_OF_MAX_TEXTURE_SIZE_mm2
					"	psrld	%%mm6,	%%mm2\n" // mm2 = (mm2 & xandmask) << xshl
					"	paddd	%%mm2,	%%mm0\n" // mm0 = boffset[2]

				//	move the four points for the second coord in mm2 and mm3
					"	movd	%%mm0,	%%eax\n"
					"	psrlq	$32,	%%mm0\n"
					"	movd	%%mm0,	%%edx\n"
					"	movl	%7,	%%ecx\n"
					"	movl	%4,	%%ebx\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm2\n"

					"	addl	%8,	%%eax\n"
					"	inc	%%eax\n"
					"	andl	%%ecx,	%%eax\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm3\n" //!!

/* situation: (low-to-hi)
	mm2 -> [ boffset		, boffset+1 		] of first coord
	mm3 -> [ boffset+xandmask+1	, boffset+xandmask+2 	] of first coord
{
	mm2(1)*=(       fx)*(0xffff-fy)
	mm2(0)*=(0xffff-fx)*(0xffff-fy)
	mm3(1)*=(       fx)*(       fy)
	mm3(0)*=(0xffff-fx)*(       fy)
}
*/

					"	pxor	%%mm7,	%%mm7\n" // nullify mm7
					"	movq	%9,	%%mm0\n" // mm0 = 0xffffffffffffffff
					"	movq	%9,	%%mm1\n" // mm1 = mm0
					"	pshufw	$0xaa, %%mm4, %%mm5\n" //extract the third word to mm5
					"	psubw	%%mm5,	%%mm0\n"
					"	pshufw	$0, %%mm4, %%mm5\n" //extract the first word to all in mm5
					"	psubw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm5\n"
					"	pmulhuw	%%mm1,	%%mm0\n"
					"	movq	%%mm2,	%%mm1\n" // save mm2 to mm1
					"	punpcklbw %%mm7,%%mm2\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7,%%mm1\n" // boffset in mm1, boffset+1 in mm2

					"	pmulhuw %%mm5,	%%mm1\n"
					"	pmulhuw %%mm0,	%%mm2\n"
					"	paddusw	%%mm1,	%%mm2\n"

					"	movq	%9,	%%mm0\n"
					"	pshufw	$0xaa, %%mm4,	%%mm5\n" // extract y0 to mm5
					"	pshufw	$0, %%mm4, %%mm1\n" // extract x0 to mm5
					"	psubw	%%mm1,	%%mm0\n"
					"	pmulhuw	%%mm5,	%%mm0\n"
					"	pmulhuw	%%mm1,	%%mm5\n"
					"	movq	%%mm3,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm3\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm1\n"

					"	pmulhuw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm3\n"
					"	paddusw	%%mm1,	%%mm2\n"
					"	paddusw	%%mm3,	%%mm2\n"
				// first point is done. Color in word fmt is in mm2.
					"	movq	%%mm2,	%%mm5\n"
					"	punpcklwd %%mm7,	%%mm2\n" // unpack words to mm2 (low)
					"	punpckhwd %%mm7,	%%mm5\n" // and mm5 (high)
					"	cvtpi2ps %%mm2,	%%xmm4\n"	 // convert mm2 to xmm4
					"	cvtpi2ps %%mm5,	%%xmm5\n"	 // convert mm5 to xmm5
					"	movlhps	%%xmm5,	%%xmm4\n"	 // get in xmm4 all components of color
					"	movaps	%%xmm6,	%%xmm5\n"
					"	shufps	$0, %%xmm6, %%xmm5\n"	// fill all xmm5 with xmm6[0]
					"	mulps	%%xmm5,	%%xmm4\n"	// multiply xmm4
				// repack to RGB entity using MMX:
					"	cvtps2pi %%xmm4,	%%mm2\n"
					"	movhlps	%%xmm4,	%%xmm5\n"
					"	cvtps2pi %%xmm5,	%%mm5\n"
					"	packssdw %%mm5,	%%mm2\n"
					"	packuswb %%mm7,	%%mm2\n"
					"	movd	%%mm2,	%0\n"
/*
						    ************	.    ************
						  **            **	.  **            **
						 *    *****       *	. *    *****       *
						*    *     *       *	.*    *     *       *
						*          *       *	.*          *       *
						*         *        *	.*         *        *
						*       **         *	.*       **         *
						 *    **     *    *	. *    **     *    *
						  **  ********  **	.  **  ********  **
						    ************	.    ************
	*/


// next point baddress is in edx... repeat the whole procedure...
					"	movl	%%edx,	%%eax\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm2\n"

					"	addl	%8,	%%eax\n"
					"	inc	%%eax\n"
					"	andl	%%ecx,	%%eax\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm3\n"

					"	movq	%9,	%%mm0\n" // mm0 = 0xffffffffffffffff
					"	movq	%9,	%%mm1\n" // mm1 = mm0
					"	pshufw	$0xff, %%mm4, %%mm5\n" //extract the third word to mm5
					"	psubw	%%mm5,	%%mm0\n"
					"	pshufw	$0x55, %%mm4, %%mm5\n" //extract the first word to all in mm5
					"	psubw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm5\n"
					"	pmulhuw	%%mm1,	%%mm0\n"
					"	movq	%%mm2,	%%mm1\n" // save mm2 to mm1
					"	punpcklbw %%mm7,%%mm2\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7,%%mm1\n" // boffset in mm1, boffset+1 in mm2

					"	pmulhuw %%mm5,	%%mm1\n"
					"	pmulhuw %%mm0,	%%mm2\n"
					"	paddusw	%%mm1,	%%mm2\n"

					"	movq	%9,	%%mm0\n"
					"	pshufw	$0xff, %%mm4,	%%mm5\n" // extract y0 to mm5
					"	pshufw	$0x55, %%mm4, %%mm1\n" // extract x0 to mm5
					"	psubw	%%mm1,	%%mm0\n"
					"	pmulhuw	%%mm5,	%%mm0\n"
					"	pmulhuw	%%mm1,	%%mm5\n"
					"	movq	%%mm3,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm3\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm1\n"

					"	pmulhuw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm3\n"
					"	paddusw	%%mm1,	%%mm2\n"
					"	paddusw	%%mm3,	%%mm2\n"
				// second point is done. Color in word fmt is in mm2.
					"	movq	%%mm2,	%%mm5\n"
					"	punpcklwd %%mm7,	%%mm2\n" // unpack words to mm2 (low)
					"	punpckhwd %%mm7,	%%mm5\n" // and mm5 (high)
					"	cvtpi2ps %%mm2,	%%xmm4\n"	 // convert mm2 to xmm4
					"	cvtpi2ps %%mm5,	%%xmm5\n"	 // convert mm5 to xmm5
					"	movlhps	%%xmm5,	%%xmm4\n"	 // get in xmm4 all components of color
					"	movaps	%%xmm6,	%%xmm5\n"
					"	shufps	$0x55,%%xmm6, %%xmm5\n"	// fill all xmm5 with xmm6[1]
					"	mulps	%%xmm5,	%%xmm4\n"	// multiply xmm4
				// repack to RGB entity using MMX:
					"	cvtps2pi %%xmm4,	%%mm2\n"
					"	movhlps	%%xmm4,	%%xmm5\n"
					"	cvtps2pi %%xmm5,	%%mm5\n"
					"	packssdw %%mm5,	%%mm2\n"
					"	packuswb %%mm7,	%%mm2\n"
					"	movd	%%mm2,	4%0\n"

/*
	    ************	    ************	    ************
	  **   ****     **	  **   ****     **	  **   ****     **
	 *    *    *      *	 *    *    *      *	 *    *    *      *
	*          *       *	*          *       *	*          *       *
	*       ***        *	*       ***        *	*       ***        *
	*          *       *	*          *       *	*          *       *
	*    *     *       *	*    *     *       *	*    *     *       *
	 *    *    *      *	 *    *    *      *	 *    *    *      *
	  **   ****     **	  **   ****     **	  **   ****     **
	    ************	    ************	    ************
	*/

// third coord:
				// restore 3rd and 4th x and y coordinate...
					"	movq	%3,	%%mm0\n"
					"	movq	8%3,	%%mm2\n"
					"	movq	%%mm0,	%%mm4\n"
					"	movq	%%mm2,	%%mm5\n"
				// shift the first copy additional 16 bits, arithmetical
					"	psrad	$16,	%%mm0\n"
					"	psrad	$16,	%%mm2\n"

					"	pslld	$16,	%%mm4\n"
					"	pslld	$16,	%%mm5\n"
					"	psrld	$17,	%%mm4\n" //mm4 &= 0xffff = FX
					"	psrld	$17,	%%mm5\n" //mm5 &= 0xffff = FY
					"	packssdw %%mm5,	%%mm4\n" //mm4 = (lo to hi) -> [ fx0 fx1 fy0 fy1 ]
					"	psllw	$1,	%%mm4\n"

					"	pand	%8,	%%mm2\n"
					"	pand	%8,	%%mm0\n" // mm0 &= xandmask
					SHIFT_LEFT_LOG2_OF_MAX_TEXTURE_SIZE_mm2
					"	psrld	%%mm6,	%%mm2\n" // mm2 = (mm2 & xandmask) << xshl
					"	paddd	%%mm2,	%%mm0\n" // mm0 = boffset[2]

				//	move the four points for the second coord in mm2 and mm3
					"	movd	%%mm0,	%%eax\n"
					"	psrlq	$32,	%%mm0\n"
					"	movd	%%mm0,	%%edx\n"
					"	movl	%7,	%%ecx\n"
					"	movl	%4,	%%ebx\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm2\n"

					"	addl	%8,	%%eax\n"
					"	inc	%%eax\n"
					"	andl	%%ecx,	%%eax\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm3\n"

					"	movq	%9,	%%mm0\n" // mm0 = 0xffffffffffffffff
					"	movq	%9,	%%mm1\n" // mm1 = mm0
					"	pshufw	$0xaa, %%mm4, %%mm5\n" //extract the third word to mm5
					"	psubw	%%mm5,	%%mm0\n"
					"	pshufw	$0, %%mm4, %%mm5\n" //extract the first word to all in mm5
					"	psubw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm5\n"
					"	pmulhuw	%%mm1,	%%mm0\n"
					"	movq	%%mm2,	%%mm1\n" // save mm2 to mm1
					"	punpcklbw %%mm7,%%mm2\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7,%%mm1\n" // boffset in mm1, boffset+1 in mm2

					"	pmulhuw %%mm5,	%%mm1\n"
					"	pmulhuw %%mm0,	%%mm2\n"
					"	paddusw	%%mm1,	%%mm2\n"

					"	movq	%9,	%%mm0\n"
					"	pshufw	$0xaa, %%mm4,	%%mm5\n" // extract y0 to mm5
					"	pshufw	$0, %%mm4, %%mm1\n" // extract x0 to mm5
					"	psubw	%%mm1,	%%mm0\n"
					"	pmulhuw	%%mm5,	%%mm0\n"
					"	pmulhuw	%%mm1,	%%mm5\n"
					"	movq	%%mm3,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm3\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm1\n"

					"	pmulhuw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm3\n"
					"	paddusw	%%mm1,	%%mm2\n"
					"	paddusw	%%mm3,	%%mm2\n"
				// third point is done. Color in word fmt is in mm2.
					"	movq	%%mm2,	%%mm5\n"
					"	punpcklwd %%mm7,	%%mm2\n" // unpack words to mm2 (low)
					"	punpckhwd %%mm7,	%%mm5\n" // and mm5 (high)
					"	cvtpi2ps %%mm2,	%%xmm4\n"	 // convert mm2 to xmm4
					"	cvtpi2ps %%mm5,	%%xmm5\n"	 // convert mm5 to xmm5
					"	movlhps	%%xmm5,	%%xmm4\n"	 // get in xmm4 all components of color
					"	movaps	%%xmm6,	%%xmm5\n"
					"	shufps	$0xaa,%%xmm6, %%xmm5\n"	// fill all xmm5 with xmm6[2]
					"	mulps	%%xmm5,	%%xmm4\n"	// multiply xmm4
				// repack to RGB entity using MMX:
					"	cvtps2pi %%xmm4,	%%mm2\n"
					"	movhlps	%%xmm4,	%%xmm5\n"
					"	cvtps2pi %%xmm5,	%%mm5\n"
					"	packssdw %%mm5,	%%mm2\n"
					"	packuswb %%mm7,	%%mm2\n"
					"	movd	%%mm2,	8%0\n"

/*
							    ************
							  **            **
							 *   ***  **      *
							*     *   *        *
							*     *   *        *
							*     *****        *
							*         *        *
							 *        *       *
							  **      **    **
							    ************
******************************************************************************
********************************        **************************************
***************************                  *********************************
********                                                          ************
********                  *****   ****   ****                     ************
********                    *       *     *                       ************
********                    *        *   *                        ************
********                    *         * *                         ************
********                  *****        *                          ************
********                                                          ************
********************************        **************************************
******************************************************************************
	*/
// FOURTH point!
// next point baddress is in edx... repeat the whole procedure...
					"	movl	%%edx,	%%eax\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm2\n"

					"	addl	%8,	%%eax\n"
					"	inc	%%eax\n"
					"	andl	%%ecx,	%%eax\n"
					"	movq	(%%ebx, %%eax, 0x4),	%%mm3\n"

					"	movq	%9,	%%mm0\n" // mm0 = 0xffffffffffffffff
					"	movq	%9,	%%mm1\n" // mm1 = mm0
					"	pshufw	$0xff, %%mm4, %%mm5\n" //extract the third word to mm5
					"	psubw	%%mm5,	%%mm0\n"
					"	pshufw	$0x55, %%mm4, %%mm5\n" //extract the first word to all in mm5
					"	psubw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm5\n"
					"	pmulhuw	%%mm1,	%%mm0\n"
					"	movq	%%mm2,	%%mm1\n" // save mm2 to mm1
					"	punpcklbw %%mm7,%%mm2\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7,%%mm1\n" // boffset in mm1, boffset+1 in mm2

					"	pmulhuw %%mm5,	%%mm1\n"
					"	pmulhuw %%mm0,	%%mm2\n"
					"	paddusw	%%mm1,	%%mm2\n"

					"	movq	%9,	%%mm0\n"
					"	pshufw	$0xff, %%mm4,	%%mm5\n" // extract y0 to mm5
					"	pshufw	$0x55, %%mm4, %%mm1\n" // extract x0 to mm5
					"	psubw	%%mm1,	%%mm0\n"
					"	pmulhuw	%%mm5,	%%mm0\n"
					"	pmulhuw	%%mm1,	%%mm5\n"
					"	movq	%%mm3,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm3\n"
					"	psrlq	$32,	%%mm1\n"
					"	punpcklbw %%mm7, %%mm1\n"

					"	pmulhuw	%%mm5,	%%mm1\n"
					"	pmulhuw	%%mm0,	%%mm3\n"
					"	paddusw	%%mm1,	%%mm2\n"
					"	paddusw	%%mm3,	%%mm2\n"
				// last point is done. Color in word fmt is in mm2.
					"	movq	%%mm2,	%%mm5\n"
					"	punpcklwd %%mm7,	%%mm2\n" // unpack words to mm2 (low)
					"	punpckhwd %%mm7,	%%mm5\n" // and mm5 (high)
					"	cvtpi2ps %%mm2,	%%xmm4\n"	 // convert mm2 to xmm4
					"	cvtpi2ps %%mm5,	%%xmm5\n"	 // convert mm5 to xmm5
					"	movlhps	%%xmm5,	%%xmm4\n"	 // get in xmm4 all components of color
					"	movaps	%%xmm6,	%%xmm5\n"
					"	shufps	$0xff,%%xmm6, %%xmm5\n"	// fill all xmm5 with xmm6[3]
					"	mulps	%%xmm5,	%%xmm4\n"	// multiply xmm4
				// repack to RGB entity using MMX:
					"	cvtps2pi %%xmm4,	%%mm2\n"
					"	movhlps	%%xmm4,	%%xmm5\n"
					"	cvtps2pi %%xmm5,	%%mm5\n"
					"	packssdw %%mm5,	%%mm2\n"
					"	packuswb %%mm7,	%%mm2\n"
					"	movd	%%mm2,	12%0\n"

				// get integer coordinates from memory:
					"	movq	%1,	%%mm0\n"
					"	movq	8%1,	%%mm1\n"
					"	movq	%2,	%%mm2\n"
					"	movq	8%2,	%%mm3\n"

				// continue interpolation:
					"	paddd	%6,	%%mm0\n"
					"	paddd	8%6,	%%mm1\n"
					"	paddd	16%6,	%%mm2\n"
					"	paddd	24%6,	%%mm3\n"
					"	addps	%%xmm2,	%%xmm0\n"
					"	addps	%%xmm3,	%%xmm1\n"
					:"=m"(*ptr), "=m"(*gx), "=m"(*gy), "=m"(*save)
					:"m"(ptex),"m"(*lightxy1),"m"(*gxx), "m"(texandmask),"m"(*xxandmask),
					 "m"(*ffs)
					:"memory", "eax", "ebx", "ecx", "edx"
					);
					ptr+=4;
				}
			} else {
			// NEAREST NEIGHBOUR FILTER VERSION:
				while (i--) {
					__asm __volatile (
				// calculate light degradation:
					"	movaps	%5,	%%xmm4\n" // load light x in xmm4
					"	movaps	0x10%5,	%%xmm5\n" // load light y in xmm5
					"	movaps	%%xmm7,	%%xmm6\n"
					"	subps	%%xmm0,	%%xmm4\n" // xmm4 = [ -current x + lightx ]   x 4
					"	subps	%%xmm1,	%%xmm5\n" // xmm5 = [ -current y + lighty ]   x 4
					"	mulps	%%xmm4,	%%xmm4\n" // xmm4 = [(-current x + lightx)^2] x 4
					"	mulps	%%xmm5,	%%xmm5\n" // xmm5 = [(-current y + lighty)^2] x 4
					"	addps	0x20%5,	%%xmm4\n" // xmm4 = xmm4 + (1, 1, 1, 1)
					"	addps	%%xmm5,	%%xmm4\n" // xmm4 = xmm4 + xmm5 = dist^2

					"	rsqrtps	%%xmm4,	%%xmm4\n" // xmm4 = 1 / dist
					"	mulps	%%xmm4,	%%xmm6\n" // xmm6 = const / dist
					"	rsqrtps %%xmm4, %%xmm4\n" // xmm4 = sqrt(dist)
					"	mulps	%%xmm4,	%%xmm6\n" // xmm6 = const / sqrt(dist)

				// save coordinates from mmx registers to memory:
					"	movq	%%mm0,	%1\n"
					"	movq	%%mm1,	8%1\n"
					"	movq	%%mm2,	%2\n"
					"	movq	%%mm3,	8%2\n"

				// shift all coords 16 positions, arithmetical
					"	psrad	$16,	%%mm0\n"
					"	psrad	$16,	%%mm1\n"
					"	psrad	$16,	%%mm2\n"
					"	psrad	$16,	%%mm3\n"
					"	movq	%7,	%%mm7\n"
				// then shift xtex positions, logical
					"	psrld	%%mm6,	%%mm0\n"
					"	psrld	%%mm6,	%%mm1\n"
					"	psrld	%%mm6,	%%mm2\n"
					"	psrld	%%mm6,	%%mm3\n"

					"	pand	%%mm7,	%%mm0\n" // mm0 &= xandmask; (mm0 -> x0 x1)
					"	pand	%%mm7,	%%mm1\n" // mm1 &= xandmask; (mm1 -> x2 x3)
					"	pand	%%mm7,	%%mm2\n" // mm2 &= xandmask; (mm2 -> y0 y1)
					"	pand	%%mm7,	%%mm3\n" // mm3 &= xandmask; (mm3 -> y2 y3)

					SHIFT_LEFT_LOG2_OF_MAX_TEXTURE_SIZE_mm2
					SHIFT_LEFT_LOG2_OF_MAX_TEXTURE_SIZE_mm3
					"	psrld	%%mm6,	%%mm2\n" // mm2 <<= xshl
					"	psrld	%%mm6,	%%mm3\n" // mm3 <<= xshl

					"	paddd	%%mm2,	%%mm0\n" // mm0 = index0 index1
					"	paddd	%%mm3,	%%mm1\n" // mm1 = index2 index3

					"	pxor	%%mm7,	%%mm7\n" // nullify mm7
					"	movl	%4,	%%ebx\n"

					"	movd	%%mm0,	%%eax\n"
					"	psrlq	$32,	%%mm0\n"
					"	movd	%%mm0,	%%edx\n"

					"	movd	(%%ebx, %%eax, 4),	%%mm2\n" // get first pixel in mm2
					"	movd	(%%ebx,	%%edx, 4),	%%mm3\n" // get second pixel in mm3

					"	punpcklbw %%mm7,	%%mm2\n"
					"	punpcklbw %%mm7,	%%mm3\n"

					"	movq	%%mm2,	%%mm4\n"
					"	movq	%%mm3,	%%mm5\n"

					"	punpcklwd %%mm7,	%%mm2\n"
					"	punpcklwd %%mm7,	%%mm3\n"
					"	punpckhwd %%mm7,	%%mm4\n"
					"	punpckhwd %%mm7,	%%mm5\n"

					"	cvtpi2ps %%mm2,	%%xmm2\n"
					"	cvtpi2ps %%mm3,	%%xmm3\n"
					"	cvtpi2ps %%mm4,	%%xmm4\n"
					"	cvtpi2ps %%mm5,	%%xmm5\n"
					"	movlhps %%xmm4,	%%xmm2\n"
					"	movlhps %%xmm5,	%%xmm3\n"
					"	movaps	%%xmm6,	%%xmm4\n"
					"	movaps	%%xmm6,	%%xmm5\n"
					"	shufps $0, %%xmm6, %%xmm4\n"
					"	shufps $0x55, %%xmm6, %%xmm5\n"
					"	mulps	%%xmm4,	%%xmm2\n"
					"	mulps	%%xmm5,	%%xmm3\n"
					"	movhlps	%%xmm2,	%%xmm4\n"
					"	movhlps	%%xmm3,	%%xmm5\n"
					"	cvtps2pi %%xmm2, %%mm2\n"
					"	cvtps2pi %%xmm3, %%mm3\n"
					"	cvtps2pi %%xmm4, %%mm4\n"
					"	cvtps2pi %%xmm5, %%mm5\n"
					"	packssdw %%mm4,	%%mm2\n"
					"	packssdw %%mm5,	%%mm3\n"
					"	packuswb %%mm7,	%%mm2\n"
					"	packuswb %%mm7,	%%mm3\n"
					// points 0 and 1 are ready, push to framebuffer:
					"	movd	%%mm2,	%0\n"
					"	movd	%%mm3,	4%0\n"
// proceed with points 2 and 3...
					"	movd	%%mm1,	%%eax\n"
					"	psrlq	$32,	%%mm1\n"
					"	movd	%%mm1,	%%edx\n"

					"	movd	(%%ebx, %%eax, 4),	%%mm2\n" // get first pixel in mm2
					"	movd	(%%ebx,	%%edx, 4),	%%mm3\n" // get second pixel in mm3

					"	punpcklbw %%mm7,	%%mm2\n"
					"	punpcklbw %%mm7,	%%mm3\n"

					"	movq	%%mm2,	%%mm4\n"
					"	movq	%%mm3,	%%mm5\n"

					"	punpcklwd %%mm7,	%%mm2\n"
					"	punpcklwd %%mm7,	%%mm3\n"
					"	punpckhwd %%mm7,	%%mm4\n"
					"	punpckhwd %%mm7,	%%mm5\n"

					"	cvtpi2ps %%mm2,	%%xmm2\n"
					"	cvtpi2ps %%mm3,	%%xmm3\n"
					"	cvtpi2ps %%mm4,	%%xmm4\n"
					"	cvtpi2ps %%mm5,	%%xmm5\n"
					"	movlhps %%xmm4,	%%xmm2\n"
					"	movlhps %%xmm5,	%%xmm3\n"
					"	movaps	%%xmm6,	%%xmm4\n"
					"	movaps	%%xmm6,	%%xmm5\n"
					"	shufps $0xaa, %%xmm6, %%xmm4\n"
					"	shufps $0xff, %%xmm6, %%xmm5\n"
					"	mulps	%%xmm4,	%%xmm2\n"
					"	mulps	%%xmm5,	%%xmm3\n"
					"	movhlps	%%xmm2,	%%xmm4\n"
					"	movhlps	%%xmm3,	%%xmm5\n"
					"	cvtps2pi %%xmm2, %%mm2\n"
					"	cvtps2pi %%xmm3, %%mm3\n"
					"	cvtps2pi %%xmm4, %%mm4\n"
					"	cvtps2pi %%xmm5, %%mm5\n"
					"	packssdw %%mm4,	%%mm2\n"
					"	packssdw %%mm5,	%%mm3\n"
					"	packuswb %%mm7,	%%mm2\n"
					"	packuswb %%mm7,	%%mm3\n"
					// points 2 and 3 are ready, push to framebuffer:
					"	movd	%%mm2,	8%0\n"
					"	movd	%%mm3,	12%0\n"

				// get integer coordinates from memory:
					"	movq	%1,	%%mm0\n"
					"	movq	8%1,	%%mm1\n"
					"	movq	%2,	%%mm2\n"
					"	movq	8%2,	%%mm3\n"

				// continue interpolation:
					"	paddd	%6,	%%mm0\n"
					"	paddd	8%6,	%%mm1\n"
					"	paddd	16%6,	%%mm2\n"
					"	paddd	24%6,	%%mm3\n"
					"	movaps	%8,	%%xmm2\n"
					"	movaps	16%8,	%%xmm3\n"
					"	addps	%%xmm2,	%%xmm0\n"
					"	addps	%%xmm3,	%%xmm1\n"

					:"=m"(*ptr), "=m"(*gx), "=m"(*gy), "=m"(*save)
					:"m"(ptex),"m"(*lightxy1),"m"(*gxx),"m"(*xxandmask), "m"(*fxyi)
					:"memory", "eax", "ebx", "edx");
					ptr += 4;
				}
			}
			prof_leave(PROF_WORK_PER_ROW);
		} else { // fill the row with zeros:
			memset(ptr, 0, 4*xr);
			ptr+=xr;
		}
		__asm __volatile ("emms");
		for (i=0;i<3;i++) tt[i]+=tti[i];
	}
}

// this is for the SSE version of the renderer, too. Renders the spheres in fbuffer and puts some
// blending data in fbuffer (zeros in fbuffers means there is no sphere for the specified pixel; see
// merge_rows in fract_asm.h for details)

// NOTE: this uses a modified code from DrawIt_P5
void render_spheres(Uint32 *fb, unsigned short *fbuffer,
		double *tt, double *ti, double *tti)
{
	int i, j, k, xr, yr, dropped, backd, ii;
	double t[3];
	double v[3];
	double lw[3];
	double opacity, bopacity;
	Uint32 fcol, bfcol;
	int foundsphere;
		prof_enter(PROF_MEMSETS);
	xr = xsize_render(xres()); yr = ysize_render(yres());
	memset(fbuffer, 0, xr*yr*sizeof(unsigned short));
	memset(RowInfo, 0, sizeof(RowInfo));
	lw[0] = lx; lw[1] = ly; lw[2] = lz;
		prof_leave(PROF_MEMSETS);
	for (j=0;j<yr;j++) {
		for (k=0;k<3;k++) t[k] = tt[k];
		for (i=0;i<xr;i++,fb++,fbuffer++,t[0]+=ti[0],t[2]+=ti[2]) {
			if (*fb) {
				dropped = *fb;
				backd = dropped >> 16;
				dropped &= 0xffff;
				v[0] = t[0]-cur[0]; v[1] = t[1] - cur[1]; v[2] = t[2] - cur[2];
				norm(v);
				if (testsphere(sp+dropped-1, v, cur)) {
					fcol = Solve3D(sp + dropped-1 , v, cur, lw, &opacity, 0);
					if (opacity > 0.99)  {
						*fb=fcol;
						*fbuffer = 65535;
					} else {
						if (backd) {
							if (testsphere(sp+backd-1, v, cur)) {
								bfcol = Solve3D(sp + backd-1, v, cur, lw, &bopacity,0);
								if (bopacity>0.99) {
									*fb = blend_sse(fcol, bfcol, opacity);
									*fbuffer = 65535;
								} else {
									*fb = blend(fcol, bfcol, opacity);
									opacity = opacity + (bopacity*(1-opacity));
									*fbuffer = (unsigned int) (opacity*65535);
								}
							} else {
								*fb=fcol;
								*fbuffer = (unsigned int) (opacity*65535);
							}
						} else {
							*fb = fcol;
							*fbuffer = (unsigned int) (opacity*65535);
						}
					}
				} else {// too bad - we'll have to iterate all the spheres to find the right one...
					foundsphere = 0;
					for (ii=spherecount-1;ii>=0;ii--)
						if (testsphere(sp + ii, v, cur)) { // sphere found...
							fcol = Solve3D(sp + ii, v, cur, lw, &opacity, 0);
							if (opacity > 0.99) {
								*fb=fcol;
								*fbuffer = 65535;
							} else {
								*fb = fcol;
								*fbuffer = (unsigned int) (opacity*65535);
							}
							foundsphere = 1;
							break;
						}
				}
				if (*fbuffer) RowInfo[j] = 1;
			}
		}
		for (k=0;k<3;k++) tt[k] += tti[k];
	}

}

static double thread_tt[3], thread_ti[3], thread_tti[3];

void thread_pass_args(double tt[3], double ti[3], double tti[3])
{
	memcpy(thread_tt, tt, sizeof(double)*3);
	memcpy(thread_ti, ti, sizeof(double)*3);
	memcpy(thread_tti, tti, sizeof(double)*3);
}

int thread_func(void *x)
{
	render_spheres(framebuffer, fbuffer, thread_tt, thread_ti, thread_tti);
	return 0;
}

void merge_buffers(Uint32 *src, Uint32 *dst, unsigned short *fbuffer)
{
	int y, xr, yr;
	xr = xsize_render(xres()); yr = ysize_render(yres());
	for (y = 0; y < yr; y++, src+=xr, dst+=xr, fbuffer+=xr) if (RowInfo[y]) {
		merge_rows(src, dst, fbuffer, xr);
	}
	//memcpy(sph, flr, xsize_render(xres()) * ysize_render(yres()) * 4);
}

/*
	This is the SSE version of the renderer (well, this routine is merely a nasty caller)
	The idea is:

	1. to render the floor very fast, using vectorized SSE code in render_floor;
	2. to render the spheres using less innermost cycle IFs, and preserving alpha data in a seperate buffer;
	3. to blend the two framebuffers into a single with merge_buffers (using MMX);
	4. for everything else use the shared code for the P5 and SSE versions (bash_preframe,
		preframe_do, postframe_do)

	The good news is, that, floor rendering is up to 5 times faster;
	The bad news are, that sphere rendering is a bit slower, and there is an additional overhead for
	framebuffer merge. Also, pixels that are rendered with the floor renderer behind the balls are
	calculated in vain, 'coz they are lost forever.
*/
void DrawIt_SSE(SDL_Surface *p, SDL_Overlay *ov)
{
	double lw[3], tt[3], ti[3], tti[3];
	double t0[3];
	int xr, yr, i;
	Uint32 *ptr;
	SDL_Thread *sphere_thread;
	long blah;

	if (UseThreads==-1) UseThreads = OptionExists("--thread");

	xr = xsize_render(xres()); yr = ysize_render(yres());

#ifdef BLUR
	ptr = blurbuffer;
#else
	ptr = floorbuffer;
#endif
	prof_enter(PROF_BASH_PREFRAME);		bash_preframe(lw, tt, ti, tti);				prof_leave(PROF_BASH_PREFRAME);
	prof_enter(PROF_PREFRAME_DO);		preframe_do(p, framebuffer, lw);			prof_leave(PROF_PREFRAME_DO);

	//save tt's
	for (i=0;i<3;i++) { t0[i] = tt[i];}

	if (!UseThreads) {
	prof_enter(PROF_RENDER_FLOOR);		render_floor(p, ov, ptr, xr, yr, tt, ti, tti);		prof_leave(PROF_RENDER_FLOOR);
	prof_enter(PROF_RENDER_SPHERES);	render_spheres(framebuffer, fbuffer, t0, ti, tti);	prof_leave(PROF_RENDER_SPHERES);
	} else {
	thread_pass_args(t0, ti, tti);
	prof_enter(PROF_RENDER_SPHERES);

	if ((sphere_thread = SDL_CreateThread(thread_func, &blah))==NULL) { // spawn child thread to render the spheres
		printf("Can't create SDL_Thread\n");
		exit(0);
	}
	used_threads = 1;
	prof_enter(PROF_RENDER_FLOOR);		render_floor(p, ov, ptr, xr, yr, tt, ti, tti);		prof_leave(PROF_RENDER_FLOOR);

	SDL_WaitThread(sphere_thread, NULL); //wait for child thread;
	prof_leave(PROF_RENDER_SPHERES);

		/*  A note on profiling : the idea of RDTSC-based profiling is totally spoiled by a multi-threaded execution:
		 * 			  you should make NO assumption about the correctness of the PROF_RENDER_SPHERES and
		 *			  PROF_RENDER_FLOOR entities in the profiling statistics in the end.
		 */
	}
	prof_enter(PROF_MERGE_BUFFERS);		merge_buffers(ptr, framebuffer, fbuffer);		prof_leave(PROF_MERGE_BUFFERS);
	prof_enter(PROF_POSTFRAME_DO);		postframe_do(p, ov, floorbuffer);			prof_leave(PROF_POSTFRAME_DO);
}


Uint32 *get_image_buffer(void)
{
	if ( sse_enabled )
		return floorbuffer;
		else return framebuffer;
}

/****************************************************************
 * This is the master function which chooses who to call to do  *
 * the actual job. Currently, the "CPU world" is divided to     *
 * SSE and non-SSE only.                                        *
 ****************************************************************/

void render_single_frame(SDL_Surface *p, SDL_Overlay *ov)
{
	if ( sse_enabled )
		DrawIt_SSE(p, ov);
		else
		DrawIt_P5(p, ov);
}
