/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *      VECTORMATH.CPP - Does anything related to vector calculation:      *
 *        - sphere pre-projection                                          *
 *        - phong shading                                                  *
 *        - physics engine                                                 *
 *                                                                         *
 ***************************************************************************/

#include "MyGlobal.h"
#include <math.h>

// this should be 45 degrees, or M_PI/4
#define A_ALPHA_OFFSET 0.78539816339745
// this should be 3/4 of A_ALPHA_OFFSET
#define A_BETA_OFFSET  0.58904862254809

#define AIRTHICKNESS_NORMAL 0.5
#define AIRTHICKNESS_HIGH   2.5
#define YI_THRESH 1

#define TIME_EPS 0.00025
// this should be the reciprocal of TIME_EPS
#define T_1_TIME_EPS 4000.0


// how much of the energy is preserved in a bounce (1 means that bounce takes no energy)
#define BOUNCE_COEFF 0.666

/*double ambient = 0.2;
double diffuse = 0.55;
double specular = 0.15;*/
extern sphere sp[];
extern int spherecount;
extern double fov;
extern double daFloor, daCeiling;

int CollDetect = 1;
int Physics = 1;
double sv_gravity = 166.66;
//double sv_gravity = 60;
double sv_air     = 3.99;

double lix[3];


#define InitMatrix(m) m[0][0]=a[0], m[1][0]=b[0], m[2][0]=-c[0], m[0][1]=0, m[1][1]=b[1], m[2][1]=-c[1], m[0][2]=a[2], m[1][2]=b[2], m[2][2]=-c[2];
#define InitMatrix2(m) m[0][0]=xv[0], m[0][1]=xv[1], m[0][2]=xv[2], m[1][0]=yv[0], m[1][1]=yv[1], m[1][2]=yv[2], m[2][0]=zv[0], m[2][1]=zv[1], m[2][2]=zv[2]

#define ReplaceCol(m, col, a) m[col][0]=a[0], m[col][1]=a[1], m[col][2]=a[2]
#define v1(a, b, c) c[0]=a[0]-b[0], c[1]=a[1]-b[1], c[2]=a[2]-b[2]
#define v2(a, b, c) c[0]=a[0]+b[0], c[1]=a[1]+b[1], c[2]=a[2]+b[2]
#define limit1(a) (a>1?1:(a))
#define arccos(x) (atan2(sqrt(1-(x)*(x)), x))
#define lift(x) ((x)*(x))
#define eq(a,b) (iabs((a)-(b))<EPS)


// calculates some grid point :
// INPUT:
// c -> camera position
// a - alpha of the point
// b - beta of the point
// alphaorig - alpha of the direction of the camera
// w - destination 3dpoint
void calc_grid_point(double *c, double a, double b, double alphaorig, double *w)
{
 w[0] = c[0] + sin(a) + sin(alphaorig)*(cos(b)-1);
 w[1] = c[1] + sin(b);
 w[2] = c[2] + cos(a) + cos(alphaorig)*(cos(b)-1);
}

// calculates the three ending points of the vector grid given a user viewpoint and a direction
void calc_grid_basics(double *c, double alpha, double beta, double w[3][3])
{
 calc_grid_point(c, alpha-A_ALPHA_OFFSET*fov*FOV_CORRECTION, beta+A_BETA_OFFSET*fov*FOV_CORRECTION, alpha, w[0]);
 calc_grid_point(c, alpha+A_ALPHA_OFFSET*fov*FOV_CORRECTION, beta+A_BETA_OFFSET*fov*FOV_CORRECTION, alpha, w[1]);
 calc_grid_point(c, alpha-A_ALPHA_OFFSET*fov*FOV_CORRECTION, beta-A_BETA_OFFSET*fov*FOV_CORRECTION, alpha, w[2]);
}

// calculates the determinant of the 3*3 matrix M.
double MatrixRoot(double M[3][3])
{
 return (M[0][0]*M[1][1]*M[2][2] + M[1][0]*M[2][1]*M[0][2] + M[0][1]*M[1][2]*M[2][0]
        -M[0][2]*M[1][1]*M[2][0] - M[2][2]*M[0][1]*M[1][0] - M[1][2]*M[2][1]*M[0][0]);
}

// rounds to the nearest integer
/*static inline int round(double x)
{int z = (int) x;
 if (x-z>0.5) z++;
 return z;
}*/

// determines where the center of the given sphere will map on the 2d screen
// works correctly with any FOV, X/Y ratio, etc.
// returns 0 if the center will be wildly of the screen (most probably when the sphere will not render at all)
int  ProjectSphere(sphere *d, double *cur, double w[3][3], int *dx, int *dy, int xres, int yres)
{
 double a[3], b[3], c[3], h[3];
 double m[3][3];
 double Dcr;
 double beta, gamma, delta;

 // init the a, b, c & h vectors
 v1(cur, w[0], h);
 v1(w[1], w[0], a);
 v1(w[2], w[0], b);
 c[0] = d->x - cur[0];
 c[1] = d->y - cur[1];
 c[2] = d->z - cur[2];

 InitMatrix(m);
 Dcr = MatrixRoot(m);
 if (iabs(Dcr)<0.01) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - zero determinant
	}

 // solve the ugly equation
 InitMatrix(m);
 ReplaceCol(m, 0, h);
 beta = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 1, h);
 gamma = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 2, h);
 delta = MatrixRoot(m) / Dcr;

 if (delta < 0) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - behind user
	}
 if (beta < -0.2 || beta > 1.2) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - out of screen
	}
 if (gamma <-0.2 || gamma> 1.2) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - out of screen
	}
 *dx = (int) round(beta * xres);
 *dy = (int) round(gamma* yres);
 return 1;
 }

// determines where the given point will map on the 2d screen
// works correctly with any FOV, X/Y ratio, etc.
// returns 0 if the point is behind the user and won't render at all (which holds even for fov = 180).
int  ProjectPoint(int *dx, int *dy, double *d, double *cur, double w[3][3], int xres, int yres)
{
 double a[3], b[3], c[3], h[3];
 double m[3][3];
 double Dcr;
 double beta, gamma, delta;

 *dx=0; *dy=0;
 // init the a, b, c & h vectors
 v1(cur, w[0], h);
 v1(w[1], w[0], a);
 v1(w[2], w[0], b);
 v1(d, cur, c);

 InitMatrix(m);
 Dcr = MatrixRoot(m);
 if (iabs(Dcr)<0.01) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - zero determinant
	}

 // solve the ugly equation
 InitMatrix(m);
 ReplaceCol(m, 0, h);
 beta = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 1, h);
 gamma = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 2, h);
 delta = MatrixRoot(m) / Dcr;

 if (delta < 0) {
 	//printf("bad delta");
 	return 0; // sphere is inprojectable - behind user
	}
 *dx = (int) round(beta * xres);
 *dy = (int) round(gamma* yres);
 return 1;
 }


double predlo[MAX_SPHERES], dlc;

void PassPre(double *p, double plc, double *LL)
{int i;
 dlc = plc;
 for (i=0;i<MAX_SPHERES;i++) predlo[i] = p[i];
 for (i=0;i<3;i++) lix[i]=LL[i];
}

// normalizes a vector at a, so that its length would become 1.
void norm(double *a)
{double t;
 t = 1 / sqrt( a[0]*a[0] + a[1]*a[1] + a[2]*a[2] );
 a[0]*=t; a[1]*=t; a[2]*=t;
}

static double ttz;

// returns 1 if the vector v (originating from cur) hits the sphere a
// returns 0 otherwise
int testsphere(sphere *a, double *v, double *cur)
{
 return ( (ttz=(sqr(a->x-(cur[0]+v[0]*a->dist)) + sqr(a->y-(cur[1]+v[1]*a->dist)) + sqr(a->z-(cur[2]+v[2]*a->dist))))<=sqr(a->d) );
}

// normalizes an angle in the 0..2*pi range
static inline double normangle(double angle)
{
 angle *= M_1_PI;
 if (angle<0)
 	angle+=2;
 if (angle>2)
 	angle-=2;
 if (angle<0 || angle >2)
 	{printf("angle error!"); fflush(stdout);}
 return angle;
}


/*************************************************************************
 * search exact dist -- find the distance to the sphere's surface        *
 *                                                                       *
 * based on the fact, that there is always a factor Q, such that         *
 *                                                                       *
 * dist(sphere, cur + Q*v) = sphere.diameter                             *
 *                                                                       *
 * it yeilds to the following quadratic equation:                        *
 *                                                                       *
 * q^2*(vx^2+vy^2+vz^2)+2*q*(fx*vx+fy*vy+fz*vz)+fx^2+fy^2+fz^2-d.d^2 = 0 *
 * where f is a vector (cur - sphere.center), d.d is sphere radius       *
 *                                                                       *
 * There should be two solutions; we are interested in the smaller       *
 * of them only.                                                         *
 *                                                                       *
 *************************************************************************/
static inline double SearchExactDist(sphere *a, double *c, double *v)
{double f[3];
 double ua, ub, uc, Dt;
 f[0]= c[0]-a->x; f[1]=c[1]-a->y; f[2]=c[2]-a->z;
 ua = sqr(v[0]) + sqr(v[1]) + sqr(v[2]);
 ub = (f[0]*v[0] + f[1]*v[1] + f[2]*v[2])*2;
 uc = sqr(f[0]) + sqr(f[1]) + sqr(f[2]) - sqr(a->d);
 Dt = sqr(ub) - 4*ua*uc;
 //if (Dt<0) { printf("SearchExactDist:error: negative determinant! it should only be positive!!!\n"); fflush(stdout); Dt=0;}

 // return the smaller root of the equation
 return ((-ub-sqrt(Dt))/(2*ua));
}


/* FindNormalVector - finds the normal of the plane, defined by the points a, b and c.

   returns 0 if the math operation is unapplicable (that is, the points are colinear)
   returns 1 and a vector in v  otherwise
*/
static inline int FindNormalVector(double *a, double *b, double *c, double *v)
{double u[3], w[3];
 v1(b, a, u);
 v1(c, a, w);
 // are u and w colinear?
 if (eq(u[0]/w[0], u[1]/w[1]) && eq(u[1]/w[1], u[2]/w[2]))
 	return 0;
 // if they are not colinear, we can calculate the cross product of u and w, which will give us the normal of the plane
 v[0] = u[1]*w[2] - u[2]*w[1];
 v[1] = u[0]*w[2] - u[2]*w[0];
 v[2] = u[0]*w[1] - u[1]*w[0];
 return 1;
}



/**********************************************************************************
 *    SOLVE3D - calculates the phong shading for a point I, which is the          *
 *              crossing point of the sphere A and the vector V, originating      *
 *              from the camera (point C), with a light source of L.              *
 *              `iteration' is used for raytracing - pass 0 when calling Solve3D  *
 *                                                                                *
 *    Returns:                                                                    *
 *             - a color, depending on the calculated phong and the r, g and b    *
 *               values of the sphere                                             *
 *             - an opacity, double between 0.0 and 1.0. Opacity is 1 anywhere    *
 *               except for the edge of the sphere, where it gradually drops.     *
 *                                                                                *
 **********************************************************************************/
Uint32 Solve3D(sphere *a, double *v, double *c, double *l, double *opacity, int iteration)
{double pff, d[3], i[3], xv[3], f[3], e[3], li[3], rt[3], doi=1/a->d, dd, dci;
 double cosbeta, cosbetalim;
 double IE, CE2;
 int rezult;
 int i2x, i2y;
 Uint32 ttt;

 if (!iteration) prof_enter(PROF_SOLVE3D);
 ttz = sqrt(ttz); // <- this assumes we've called testsphere before Solve3D

 // check if we're close to the edge:
 if((pff=ttz/a->d)>0.92) *opacity = 1-(pff-0.92)/0.08;
 		else *opacity = 1;
 d[0] = a->x; d[1] = a->y; d[2] = a->z;

 // find the distance from the viewer's point to the sphere
 dci = SearchExactDist(a, c, v);
 i[0] = c[0] + dci * v[0]; // calculate the point I where the ray from the user crosses the sphere's surface
 i[1] = c[1] + dci * v[1];
 i[2] = c[2] + dci * v[2];

 // xv is a vector, originating from D, pointing towards I
 v1(i, d, xv);

 xv[0]*=doi; xv[1]*=doi; xv[2]*=doi; //normalise the vector xv

 // li is a vector, originating from I, pointing towards L
 v1(l, i, li);

 // find the dot product of li and xv
 dd = li[0]*xv[0] + li[1]*xv[1] + li[2]*xv[2];

 // point f is the projection of L onto the vector xv
 f[0] = i[0] + dd*xv[0];
 f[1] = i[1] + dd*xv[1];
 f[2] = i[2] + dd*xv[2];

 /**********************************
  a little bit of explanation here:

  we're working on a plane, defined by
  the points L and I and the vector xv.

  Point E is on the same plane and
  is a mirror image of L, with a mirroring
  line xv.

  **********************************/
 e[0] = l[0] + 2*(f[0] - l[0]);
 e[1] = l[1] + 2*(f[1] - l[1]);
 e[2] = l[2] + 2*(f[2] - l[2]);

 // here in E we have the 3dpoint of the reflection of L from the ball (the light ray goes L -> I -> E

 // to calculate phong shading we need the cosine of the angle CIE

 CE2 = sqr(c[0]-e[0]) + sqr(c[1] - e[1]) + sqr(c[2] - e[2]);
 IE = pith(i, e);
 cosbeta = limit1((sqr(dci) + sqr(IE) - CE2) / (2*dci*IE));
 //if (cosbeta<-1) { printf("Bad beta - %.5lf\n"); fflush(stdout); }
 cosbetalim = sqrt( (1 + cosbeta) * 0.5 );
 // the original Phong Shading formulae:
 //cosbetalim += 0.005 / (cosbetalim+0.05);
 cosbetalim += 0.2;
 rezult = (int) ((ambient + lift(cosbeta)*specular + cosbetalim * diffuse)*65536.0);
 if (rezult>65536) rezult = 65536;
 if (a->flags & MAPPED) {
 	i2x = (int) (0.5*normangle(atan2(-i[0]+a->x, i[2]-a->z))*a->tex->x);
	i2y = (int) (0.5*((a->y+a->d-i[1])/a->d)*a->tex->y);
	rezult = multiplycolor(a->tex->data[(i2x + i2y * a->tex->x)%(a->tex->x*a->tex->y)], rezult);
 	}
 else rezult=(((a->b * rezult)>>16) + (((a->g * rezult)>>16)<<8) + ((a->r * rezult) & 0xff0000));
 if (!iteration) prof_leave(PROF_SOLVE3D);
 if (! (a->flags & RAYTRACED)) return rezult;
 	else {
		if (!iteration) prof_enter(PROF_RAYTRACE);
	 // li is a vector, originating at I, pointing towards C
 	v1(c, i, li);

 	// find the dot product of li and xv
 	dd = li[0]*xv[0] + li[1]*xv[1] + li[2]*xv[2];

 	// point f is the projection of C onto the vector xv
 	f[0] = i[0] + dd*xv[0];
 	f[1] = i[1] + dd*xv[1];
 	f[2] = i[2] + dd*xv[2];

 	e[0] = c[0] + 2*(f[0]-c[0]);
 	e[1] = c[1] + 2*(f[1]-c[1]);
 	e[2] = c[2] + 2*(f[2]-c[2]);

	v1(e, i, rt);

	ttt = blend(Raytrace(i,rt, a->flags & RECURSIVE, iteration+1), rezult, a->refl);
	if (!iteration) prof_leave(PROF_RAYTRACE);
	return ttt;
	}
}


// returns the color of the ray v, originating from `cur'.
// `iteration' is a counter which prevents a ray from entering an endless cycle (very rare, but possible!)
// `recursive' is a boolean flag: if 0, the procedure won't check if the ray hits any sphere.
//                                if 1 and a sphere is being hit the same procedure is recursively called for the ray
//                                that goes out of the sphere then.
Uint32 Raytrace(double *cur, double *v, int recursive, int iteration)
{double dp, ydist,scalefactor;
 double cx, cy, dist, xx[3], mind;
 sphere *z;
 int i,j;
 if (iteration > MAX_RAYTRACE_ITERATIONS)
 	return RAYTRACE_BLEND_COLOR;
 if (recursive) {
 	norm(v);
	mind = 999999999.0;
	z = NULL;
	for (i=0;i<spherecount;i++) {
		//printf("."); fflush(stdout);
		dist = pith0(sp[i].x, sp[i].y, sp[i].z, cur[0], cur[1], cur[2]);
		if (dist<mind) {
			for (j=0;j<3;j++) xx[j]=cur[j]+v[j]*dist;
			if (pith0(sp[i].x, sp[i].y, sp[i].z, xx[0], xx[1], xx[2]) < sp[i].d) {
				z = sp+i;
				mind = dist;
				}
			}
		}
	if (z!=NULL) {
		return Solve3D(z, v, cur, lix, &dp, iteration+1);
		}
 	}
 ydist = fabs(v[1]);
 if (ydist<DST_THRESHOLD)
 	return RAYTRACE_BLEND_COLOR;
 // if (floor) {
 dp = (v[1]<0?(cur[1]):(200-cur[1]));
 scalefactor = dp/ydist;
 cx = cur[0] + v[0]*scalefactor;
 cy = cur[2] + v[2]*scalefactor;
 return texture_handle((int) cx, (int) cy);
 //}
}


// performs time calculation.
//
//  time -> frame difference, in seconds
//  resistance -> air resistance, see below
//  value -> current speed

// resistance is:
// 0-1 for no resistance
// 2 - ball slows down to 50% of its speed for a second
// 9 - ball slows down to 1/9 of its speed for a second
// etc. etc.

// dir is direction. if 1 the direction is forward (air resisance is applied), otherwise is backward (resistance is UNapplied)
double time_calc(double time, double resistance, double value, int dir)
{
    double res;
    if (resistance <= 1.0) return value;
    if (dir == 1)
    	res = value / exp(log(resistance) * time);
	else
	res = value * exp(log(resistance) * time);
    return res;
}

// adds the gravity to the Y component of the vector
void apply_gravity(sphere *a, double t)
{
 if (a->flags & GRAVITY) {
 	a->yi -= sv_gravity*(t-a->time);
 }
}

// mask is a bitmask: which components will the air be applied to
// Use APPLY_ALL to mean all components
// dir is direction: if 1, the air resistance is applied, otherwise it is undone ("un-applied")
void apply_air(sphere *a, double t, int mask, int dir)
{
 double multiplyer;
 if (a->flags & AIR) {
 	if (a->flags & GRAVITY) multiplyer = sv_air * AIRTHICKNESS_NORMAL;
 			 else   multiplyer = sv_air * AIRTHICKNESS_HIGH;

 	if (mask & APPLY_X) a->xi = time_calc((t-a->time) , multiplyer, a->xi, dir);
 	if (mask & APPLY_Y) a->yi = time_calc((t-a->time) , multiplyer, a->yi, dir);
 	if (mask & APPLY_Z) a->zi = time_calc((t-a->time) , multiplyer, a->zi, dir);
 }
}


double get_hit_time(double a, double b, double c)
{
	double Dt;
	Dt = b*b - 4*a*c;
	if (Dt<0) {
		//printf("Bug: Determinant in get_hit_time < 0!\n");
		Dt = 0;
	}
	return (-b-sqrt(Dt)) / (2*a);
}

/********************************
 * void advance:                *
 *      input:                  *
 *      *a - sphere to advance; *
 *      t - sync time.          *
 *------------------------------*
 * moves a sphere depending on  *
 * its movement vector and the  *
 * time remaining to the sync   *
 * time. Checks if the sphere   *
 * hits the floor or the        *
 * ceiling - if so, a bounce is *
 * generated.                   *
 ********************************/

void advance(sphere *a, double t)
{double ytoadd, x;
 int bModif = 0;
 a->x += a->xi * (t-a->time);
 a->z += a->zi * (t-a->time);

 ytoadd = a->yi * (t-a->time);
 if (a->flags & GRAVITY)
 	if (a->y + ytoadd < daFloor + a->d) {// check for floor bounce
 		bModif = 1;
		/* so we found out that the bounce occurs. So we should make some correction to gravity, because
		 * `apply_gravity' assumes the acceleration towards the floor occurs all the way in the same direction.
		 * This ain't true. We should apply the gravity prior to the bounce, then change the sign of the Y component
		 * of the vector, and finally apply the needed amount of gravity after the bounce. (Thinking how gravity
		 * applies to the movement vector, it applies with different signs before and after the bounce)
		 */

		// first, we un-do the gravity, which has been previously applied by `apply_gravity'//
		// before that, undo air resistance:
		apply_air(a,t, APPLY_Y, -1);
		// undo gravity
		a->yi+= sv_gravity * (t - a->time);

		// calculate when it hits the floor.
		x = a->time + get_hit_time(-sv_gravity, a->yi, a->y - a->d);

		if (x>t) {
			//printf("Bug: bounce time was outside the sync time!\n");
			x = t;
			} else if (x<a->time) {
			//printf("Bug: bounce time was before start time!\n");
			x = a->time;
			}
		a->yi = -(a->yi - (x-a->time) * sv_gravity) * BOUNCE_COEFF;
		a->y = a->d + daFloor + (t-x)*a->yi;
		a->yi -= (t-x)*sv_gravity;

		if (a->yi<YI_THRESH) {
		// if the ball is slowed down enough, we may consider it stopped jumping and avoid
		// any further bounce checks
			a->yi=0;
			a->y = a->d+daFloor;
			a->flags &= ~GRAVITY;     // don't do gravity anymore.
			}
 		}
 if (a->flags & GRAVITY)
 	if (a->y + ytoadd > daCeiling - a->d) {// check for ceiling bounce
	// we use much simplier routine, since a ball cannot roll on the ceiling, right
	// (except sv_gravity is below zero, but ... :)
 		bModif = 1;
		ytoadd = (a->y + ytoadd) - (daCeiling - a->d);
		a->y = daCeiling - a->d - ytoadd;
		a->yi = -a->yi * BOUNCE_COEFF;
 		}
 if (!bModif)
	a->y += ytoadd;
 a->time = t;
}

// calculates the distance between a and b at time t
static inline double time_dist(sphere *a, sphere *b, double t)
{double t1, t2;
 t1 = t - a->time;
 t2 = t - b->time;
 return (sqrt(sqr(a->x+a->xi*t1-b->x-b->xi*t2)+sqr(a->y+a->yi*t1-b->y-b->yi*t2)+sqr(a->z+a->zi*t1-b->z-b->zi*t2)));
}

// checks if two spheres collide until time `t'.
// if collision occurs it returns the exact collision time. Else a value of -1 is returned
double collide (sphere *a, sphere *b, double t)
{double l, r, m, d0, d1;
 r = t;
 l = (a->time>b->time?a->time:b->time); // get the latter of the a's and b's times
 d0 = time_dist(a, b, l);
 while (r-l>TIME_EPS) {
 	m = (l+r)/2.0;
	d0 = time_dist(a, b, m);
	d1 = time_dist(a, b, m-TIME_EPS);
	if (d1 < d0)
		r = m;
	else
		l = m;
 	}
 if (d0 > a->d+b->d)
 	return -1; // the spheres do not collide
 //
#ifdef PRINT_COLLISIONS
 printf("Collision detected!\nA:\n", t); 
 print_sphere_info(stdout, a);
 printf("B:\n");
 print_sphere_info(stdout, b);
 printf("--> lowest time : %.6lf\n", m);
 fflush(stdout);
#endif
 l = (a->time>b->time?a->time:b->time); // get the latter of the a's and b's times

 do {
 	m = (l+r) / 2.0;
	d0 = time_dist(a, b, m);
	if (d0 > a->d + b->d)
		l = m;
	else
		r = m;
 	} while (r-l > TIME_EPS);
#ifdef PRINT_COLLISIONS
 printf("--> collision time : %.6lf\n", l); fflush(stdout);
#endif
 return l;
}

// assumes that a and b collide at time t. Upon that assumption it calculates the new motion vectors for a and b
void ProcessIncident(sphere *a, sphere *b, double t)
{double strength, d1;
 a->x += a->xi * (t- a->time); a->y += a->yi * (t- a->time);  a->z += a->zi * (t- a->time);
 b->x += b->xi * (t- b->time); b->y += b->yi * (t- b->time);  b->z += b->zi * (t- b->time);
 a->time = t;
 b->time = t;
 d1 = time_dist(a, b, t);
 strength = d1 - time_dist(a, b, t+TIME_EPS);
 d1 *= 0.5;
 if (strength<0) return; // the hit is too weak
 if (!(a->flags & STATIC)) {
 	a->xi += (a->x - b->x)/d1 * b->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	a->yi += (a->y - b->y)/d1 * b->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	a->zi += (a->z - b->z)/d1 * b->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
	}
 if (!(b->flags & STATIC)) {
 	b->xi += (b->x - a->x)/d1 * a->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	b->yi += (b->y - a->y)/d1 * a->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	b->zi += (b->z - a->z)/d1 * a->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
	}
 // custom onHit modification goes here:
 if (!(a->flags & STATIC)) {
 	a->flags &= ~ANIMATED;
	a->flags |= GRAVITY|AIR;
 	}
 if (!(b->flags & STATIC)) {
 	b->flags &= ~ANIMATED;
	b->flags |= GRAVITY|AIR;
 	}
}


/*void modify(int which, double amount)
{
 if (which==0) diffuse+=amount;
          else specular+=amount;
}*/

void vectormath_close(void)
{
 //printf("diffuse: %.2lf, specular: %.2lf\n", diffuse, specular);
}
